/*
*-----------------------------------------------------------------------------
*	file	: nncreat.c
*	desc	: create the neocognitron datastructures
*	by	: Frank Schnorrenberg
*	date	: 04-13-92
*-----------------------------------------------------------------------------
*/
#include	<stdio.h>
#include	<stddef.h>
#include	<stdlib.h>
#include	<values.h>
#include	<math.h>

#include	"nntype.h"
#include	"nnerror.h"
#include	"nncreat.h"

/*
*=============================================================================
*	funct:	nn_creat
*	dscpt:	creat a NET object
*	given:
*
*	retrn:	allocated NET
*=============================================================================
*/
NET * nn_creat()
{
	int	i, flag;
	NET	*nntmp;
	
	/* malloc the NET struct */

	if ((nntmp= (NET *)malloc(sizeof(NET))) == NULL){
		/* Failure */
		error( NNMALLOC_NET );

	} else {	/* Success */

		/*	malloc the STAGE *'s */

		DimNet(nntmp) = N_STAGES;
		if ((nntmp->stage = (STAGE **)malloc(sizeof(STAGE *) * DimNet(nntmp))) == NULL){
			/* Failure */
			error( NNMALLOC_NET );

		} else {	/* Success */
			/*
			* Create the stages including the input layer
			*/
			flag = 1;
			i = 0;
			while (flag == 1) {
				Stage(nntmp,i) = st_creat(i);
				i = i+1;
				if (i > N_STAGES - 1) flag = 0;
			}

			return (nntmp);
		}
	}
}

/*
*=============================================================================
*	funct:	v_creat 
*	dscpt:	create an allocated VECTOR object
*	given:	dim = dimension of the vector
*	retrn:	allocated VECTOR object
*=============================================================================
*/
VECTOR         *
v_creat(dim)
	int             dim;
{
	VECTOR         *vtmp;
	int             i, j;


	if (DEBUG_2 == 1) {
		printf("Entering MALLOC for VECTOR objects\n\n");
	}
	if ((vtmp = (VECTOR *) malloc(sizeof(VECTOR))) == NULL) {
		error(NNMALLOC_VEC);
	} else {
		DimVect(vtmp) = dim;
		/*
		 * In case of dummy wgts I want to have the objects but no
		 * further components allocated
		 */
		if (dim != 0) {
			/* The weights should be arranged in 2D */

			if (DEBUG_2 == 1) {
				printf("Entering wgt array MALLOC\n\n");
			}
			if ((Vect(vtmp) = (double **) malloc(sizeof(double *) * dim)) == NULL) {
				error(NNMALLOC_VEC);
			} else {
				for (i = 0; i < dim; i++) {
					if ((vtmp->vect[i] = (double *) malloc(sizeof(double) * dim)) == NULL) {
						error(NNMALLOC_VEC);
					}
				}

				if (DEBUG_2 == 1) {
					printf("2D array for wgts has been allocated at base address %d\n\n", &(vtmp->vect[0]));
				}
				v_fill(vtmp, 0.0);

				if (DEBUG_2 == 1) {
					printf("Allocation of wgt array for unit done\n\n");
				}
				return (vtmp);
			}
		} else {
			Vect(vtmp) = NULL;
			return (vtmp);
		}
	}
}

/*
*=============================================================================
*	funct:	v_fill 
*	dscpt:	fill all dim of a vector with a value
*	given:	v = vector, m = value
*	retrn:	v
*=============================================================================
*/
VECTOR * v_fill( v, m )
VECTOR	*v;
double	m;
{
	int	i, j;

	for (i=0; i<DimVect(v); i++) {
		for (j=0; j<DimVect(v); j++) {
			Vi(v,i,j) = m;
		}
	}
	return (v);
}

/*
*=============================================================================
*	funct:	v_rand
*	dscpt:	fill a vector with random value (0 - 1) 
*
*	given:	v = vector
*	retrn:	v
*=============================================================================
*/
VECTOR *v_rand( v )
VECTOR	*v;
{
	int	i, j=NULL;

	for (i=0; i<DimVect(v); i++) {
		for (i=0; i<DimVect(v); i++) {
			Vi(v,i,j) = rand();
			/* RAND_MAX IS NON_EXISTANT ON THIS SUN */
			/* i(v,i,j) = (double) (rand()%RAND_MAX) / (double) (RAND_MAX - 1); */
		}
	}
	return (v);
}


/*
*=============================================================================
*	funct:	u_creat
*	dscpt:	create an allocated UNIT, no argument needed since the weights
*		are going to be assigned to the plane
*	[given:	dimwgtvect = dimension of the unit's weight vector]
*	retrn:	allocated UNIT object
*=============================================================================
*/
/*
UNIT * u_creat()
{
	int	i;
	UNIT	*utmp;

	if ((utmp=(UNIT *)malloc(sizeof(UNIT))) == NULL)
	{
		error(NNMALLOC_UNIT);
	} else {
 
		Out(utmp) = 0.0;
		Net(utmp) = 0.0;
		Spec(utmp) = 0.0;
*/

		/* allocation successful */
/*		return (utmp);
	}
}
*/

/*
*=============================================================================
*	funct:	pl_creat
*	dscpt:	create an allocated PLANE object
*	given:	dimplane =	number of units in this plane
*		dimwgtvect =	dimension of the weight vector of each unit
*		dimnumrfld = 	number of receptive fields of the 'seed' unit
*				(so to say)
*	retrn:	allocated PLANE object
*=============================================================================
*/
PLANE * pl_creat( dimLayer, which, type, dimplane, dimnumrfld, dimwgtvect )
int	dimLayer; /*	what is the dimension of the layer,
			where this plane will belong to */
int	which;	/* which plane am i */
int	type; 
int	dimplane;
int	dimnumrfld;
int	dimwgtvect;
{
	PLANE	*pltmp;
	UNIT	*untmp;
	int	i, j, k;

	if (DEBUG_1 == 1){
		printf("Allocating plane.\n\n");
	}
	if ((pltmp= (PLANE *)malloc(sizeof(PLANE))) == NULL)
	{
		error(NNMALLOC_PLANE);
	} else {
		/*	In case of dummy planes I want to have the objects but no further
			units or weights allocated */
		DimPlane(pltmp)	= dimplane;
		DimRfld(pltmp)	= dimnumrfld;

		if (DEBUG_1 == 1){
			printf("Allocating units in plane.\n\n");
		}
		if (dimplane != 0) {
			if ((pltmp->unit= (UNIT **)malloc(sizeof(UNIT *) * dimplane)) == NULL){
				error(NNMALLOC_PLANE);
			} else {
				/*	Planes are represented as 2D objects: It is a 2D array
					of pointers to unit-objects */
				for (i=0; i<dimplane; i++) {

 					if (DEBUG_1 == 1){
						printf("Allocated Unit NOT YET assigned.\n\n");
						printf("Address of %d's row vector of plane: %d\n\n", i, &(pltmp->unit[i]));
					}

					if ( (pltmp->unit[i] = (UNIT *)malloc(dimplane * sizeof(UNIT))) == NULL){
						error(NNMALLOC_UNIT);
					}

					if (DEBUG_1 == 1){
						printf("Allocated Unit assigned.\n\n");
					}
				}

				/* Initialize */
				for (i=0; i<dimplane; i++) {
					for(j=0; j<dimplane; j++){
						Out(pltmp->unit[i][j])	= 0.0;
						Net(pltmp->unit[i][j])	= -1.0;
						Spec(pltmp->unit[i][j])	= -1.0;
					}
				}

				pltmp->dimwgt	= dimnumrfld;
				pltmp->dim	= dimwgtvect;

			}

				
			/*	Allocate weightvectors: I allocate just one set of weight vectors for each
				receptive field per plane in accordance with the papers. This set of weights
				will be assigned to the plane structure */
			if (dimnumrfld != 0) {

				if (DEBUG_2 == 1){
					printf("Entering weightvector allocation\n\n");
				}

				if ((pltmp->wgtvect= (VECTOR **)malloc(sizeof(VECTOR *) * dimnumrfld)) == NULL) {
					error(NNMALLOC_WGTVEC);

 				if (DEBUG_2 == 1){
					printf("pltmp->wgtvect[]' pointer array created at address %d.\n\n", &(pltmp->wgtvect[0]));
				}

				} else {
					/* If the V cells should be treated specially, the loop can
					be terminated early and special statements can control the allocation
					of memory for the V cells. The best solution would be to have 
					a global index array, filled with the sizes of the receptive fields */
					for (i=0; i<dimnumrfld; i++) {

 						if (DEBUG_2 == 1){
							printf("Entering the creation of the row vectors of the wgt array\n\n");
						}

						vWeight(pltmp, i) = v_creat( dimwgtvect );

 						if (DEBUG_2 == 1){
							printf("pltmp->wgtvect[i]' row created at address %d.\n\n", &(pltmp->wgtvect[i]));
						}

					}
				}
			}

			/* 	Allocate the array for the inhibitory connections if the object
				belongs to a S layer. Allocate only if the object is the last
				plane in the indexing of the layer. */

			if ( (which == dimLayer -1) && (type == 0) ){
				if( (pltmp->inhib = (double ***)malloc(sizeof(double *) * dimplane )) == NULL){
					error( NNMALLOC_INHIBERR );
				} else {
					for(i=0; i< dimplane; i++){
				  		if ( (pltmp->inhib[i] = (double **)malloc(sizeof(double *) * dimplane )) == NULL){
							error( NNMALLOC_INHIBERR );
						}
					}
					for(i=0; i< dimplane; i++){
						for(j=0; j< dimplane; j++){
							if ( (pltmp->inhib[i][j] = (double *)malloc(sizeof(double) * (dimLayer -1))) == NULL){
								error( NNMALLOC_INHIBERR );
							}
						}
					}

					/* initialize the inhibitory connections in the S layer */
					for (i=0; i<dimplane; i++){
						for (j=0;j<dimplane; j++){
							for (k=0; k<(dimLayer - 1); k++){
								pltmp->inhib[i][j][k] = 0.0;
							}
						}
					}
				}
			}

			/* allocation successful */
			pltmp->init = 0;
			return (pltmp);

		} else {
			pltmp->unit=NULL;
			return (pltmp);
		}
	}
}


/*
*=============================================================================
*	funct:	l_creat
*	dscpt:	create an allocated LAYER object
*	given:	type = S or C layer (0 or 1)
*		dimlayer = number of planes in this layer
*		dimplane = number of units in the planes in the layer
*		dimnumrfld = number of receptive fields of the 'seed' unit (so to say)
*		dimwgtvect = dimension of the weight vector of each unit
*	retrn:	allocated LAYER object
*=============================================================================
*/
LAYER * l_creat( type, dimlayer, dimplane, dimnumrfld, dimwgtvect )
int	type; /* 0 = S-layer, 1= C-layer */
int	dimlayer;
int	dimplane;
int	dimnumrfld;
int	dimwgtvect;
{
	LAYER	*ltmp;
	int	i;

	if (DEBUG_1 == 1){
		printf("Allocating Layer.\n\n");
	}
	if ((ltmp= (LAYER *)malloc(sizeof(LAYER))) == NULL) {
		error(NNMALLOC_LAY);
	} else {
		DimLayer(ltmp) = dimlayer;
		/* In case of dummy layers I want to have the objects but no further
		planes allocated */
		if (dimlayer != 0) {

			if (DEBUG_1 == 1){
				printf("Allocating planes in layer structure.\n\n");
			}

			if ((ltmp->plane= (PLANE **)malloc(sizeof(PLANE *) * dimlayer)) == NULL) {
				error(NNMALLOC_LAY);
			} else {
				for (i=0; i<dimlayer; i++)
					{
					if (DEBUG_1 == 1){
						printf("Allocating plane objects to plane pointers in layer structure.\n\n");
					}
					Plane(ltmp,i) = pl_creat(dimlayer, i, type, dimplane, dimnumrfld, dimwgtvect );
				}

				if (DEBUG_1 == 1){
					printf("Allocating layer done.\n\n");
				}


				/* allocation successful */
				return (ltmp);
			}
		} else {
			ltmp->plane=NULL;
			return (ltmp);
		}
	}
}


/*
*=============================================================================
*	funct:	st_creat
*	dscpt:	create an allocated STAGE object
*	given:	dimSlayer = number of planes in the S-layer
		dimClayer = number of planes in the C-layer
		dimSplane = number of units in the S planes
		dimCplane = number of units in the C planes
*		dimSwgtvect = dimension of the weight vector of each unit in
				each plane in each S layer
*		dimCwgtvect = dimension of the weight vector of each unit in
				each plane in each C layer
*	retrn:	allocated STAGE object
*=============================================================================
*/
STAGE * st_creat( which )
int	which;
{
	int	dimstage;/* # of layers per stage */
	int	dimSlayer, dimClayer;	/* # of planes per layer */
	int	dimSplane, dimCplane;	/* # of units to be assigned to the plane */
	int	dimSRfld, dimCRfld;	/* # of receptive fields of the planes */
	int	dimSwgtvect, dimCwgtvect;/* size of the receptive fields */
	STAGE	*sttmp;
	int	i;
	int	type=-1;

	/*	Set up of the initialization parameters:
		The dimensions given are ment to be taken square to get the 
		total amount of e.g. units in a plane. Since we are using square
		receptive fields and planes, only the size of one dimension
		is given, to make the instantiation easier */
	switch(which)
		{
		/*	case 0 is a special case: how to handle the weights
			coming from the input units? I plan to just disregard them
			for the input plane and to only set the corresponding Vc cells */
		case 0:
			dimstage	=	N_LAYERS_0;
			dimSlayer	=	0;
			dimClayer	=	N_PLANES_C0; 	/* number of planes on C layer */
			dimSplane	=	0;
			dimCplane	=	S_PLANES_C0;	/* dimension of a C plane */

			/*	# of receptive fields for S planes (actually the units, but weights
				are local to the plane */
			dimSRfld	=	Num_RcptFld_S0;	
			dimCRfld	=	Num_RcptFld_C0;

			dimSwgtvect	=	DimWgtVec_S0;	/* dimension of receptive field */
			dimCwgtvect	=	DimWgtVec_C0;

			break;
		case 1:
			dimstage	=	N_LAYERS_Rest;
			dimSlayer	=	N_PLANES_S1;
			dimClayer	=	N_PLANES_C1;
			dimSplane	=	S_PLANES_S1;
			dimCplane	=	S_PLANES_C1;
			dimSRfld	=	Num_RcptFld_S1;
			dimCRfld	=	Num_RcptFld_C1;
			dimSwgtvect	=	DimWgtVec_S1;
			dimCwgtvect	=	DimWgtVec_C1;

			break;
		case 2:
			dimstage	=	N_LAYERS_Rest;
			dimSlayer	=	N_PLANES_S2;
			dimClayer	=	N_PLANES_C2;
			dimSplane	=	S_PLANES_S2;
			dimCplane	=	S_PLANES_C2;
			dimSRfld	=	Num_RcptFld_S2;
			dimCRfld	=	Num_RcptFld_C2;
			dimSwgtvect	=	DimWgtVec_S2;
			dimCwgtvect	=	DimWgtVec_C2;

			break;
		case 3:
			dimstage	=	N_LAYERS_Rest;
			dimSlayer	=	N_PLANES_S3;
			dimClayer	=	N_PLANES_C3;
			dimSplane	=	S_PLANES_S3;
			dimCplane	=	S_PLANES_C3;
			dimSRfld	=	Num_RcptFld_S3;
			dimCRfld	=	Num_RcptFld_C3;
			dimSwgtvect	=	DimWgtVec_S3;
			dimCwgtvect	=	DimWgtVec_C3;

			break;
		case 4:
			dimstage	=	N_LAYERS_Rest;
			dimSlayer	=	N_PLANES_S4;
			dimClayer	=	N_PLANES_C4;
			dimSplane	=	S_PLANES_S4;
			dimCplane	=	S_PLANES_C4;
			dimSRfld	=	Num_RcptFld_S4;
			dimCRfld	=	Num_RcptFld_C4;
			dimSwgtvect	=	DimWgtVec_S4;
			dimCwgtvect	=	DimWgtVec_C4;

			break;
		default:
			printf("ERROR:	You've gone to far. Too many stages indicated\n\n");
		}
	/*	END: Set up of the initialization parameters */



	if ((sttmp= (STAGE *)malloc(sizeof(STAGE))) == NULL) {
		error( NNMALLOC_STA );
	} else {
		if (DEBUG_1 == 1){
			printf("Allocating layer array in stage structure.\n\n");
		}
		DimStage(sttmp) = dimstage;
		if (	((sttmp->layerS= (LAYER *)malloc(sizeof(LAYER *) )) == NULL) || 
			((sttmp->layerC= (LAYER *)malloc(sizeof(LAYER *) )) == NULL) ) {
			error(NNMALLOC_STA);
		} else {
			/*if (DEBUG_1 == 1){
				printf("Allocating c_list in stage structure.\n\n");
			}

			if ( ( sttmp->c_list = (double **) (malloc(sizeof(double *) * dimSlayer)) ) == NULL ) {
				error( NN_C_VALERR);
			} else {*/
			/* The parameters are passed from this level down to the level 
			they are used, like the 'dimSRfld' parameter, which determines
			the number of receptive fields of a S layer unit in the specified
			setting (Attention: The units in a plane share the weights).
			One can see, that even though there are many parameters which
			are only used later, the current specs determine which scope they
			have. It is done that way because, even though the parameters
			have to be specified now, a certain finetuning is possible later
			on, for which cases no general rule can be made up (e.g. clustering
			of planes) */

			if (DEBUG_1 == 1){
				printf("Allocating S-layer in stage structure.\n\n");
			}
			LayerS(sttmp) = l_creat( 0 , dimSlayer, dimSplane, dimSRfld, dimSwgtvect );

			if (DEBUG_1 == 1){
				printf("Allocating C-layer in stage structure.\n\n");
			}

			LayerC(sttmp) = l_creat( 1 , dimClayer, dimCplane, dimCRfld, dimCwgtvect );

			/* allocation successful */
			return (sttmp);
		}
	}
}
