/*
*-----------------------------------------------------------------------------
*	file	: nnio.c
*	desc	: doing the I/O part of the program
*	by	: Frank Schnorrenberg
*	date	: 05-02-92
*
*
*	comment:
*	
*-----------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>	

#include "nntype.h"
#include "nnerror.h"
#include "nncreat.h"
#include "neov.h"
#include "cmd_pars.h"

int	sample=0;	/*	number of samples to be read from the input
				file */
int	maxiter;	/*	maximal number of iterations */
int	rep;		/*	number of repetitions of a pattern before
				switch to next */
int	use;

char	tname[128], wgtname[128], specname[128], outname[128];
double	***inputvect;

double	r[N_STAGES];
double	q[N_STAGES];
double	alpha=0.0;

/*
long	time_start;
long	time_end;
*/

struct{
	double	thresh;
	int	stage;
	int	layer;
	int	plane;
	int	flag;
} disp[3];


/*double dif_time()
{

return((double)(time() - time_start));

}

double * min_sec(time_arr)
double	*time_arr;
{

	time_arr[0] = (double) ( (int)( dif_time() / 60.0) );
	time_arr[1] = (double) ( (int)(dif_time())%60 );
	
	return(time_arr);
}
*/

/* This gives a short description of how to use the program */
void	usage( )
{
	printf( "\n\n\nWhat?	%s- the executable\nWhy?	%s\nWho?	%s\n", PROGNAME, VERSION, AUTHOR );
	printf( "When?	%s\n\n", DATE);
	printf( "Description:	Simulation of Fukushima's Neocognitron\n\n");
	printf( "Usage:\n	%s -fin=<fn> -samp=# -fwgt=<fn> -use=# -fspec=<fn> -fout=<fn> -maxiter=# -reps=#\n\n", PROGNAME);
	printf( "Example:	" );
	printf( "Let the network read and classify 10 sample pattern\n\n");
	printf( "	%s -fin=myinput.pat -samp=10 -fwgt=wgts.dat -fspec=specs.dat -maxiter=100 -reps=5\n\n", PROGNAME );
	printf( "where\n\n" );
	printf( "	-fin=		name of train file containing the input pattern\n" );
	printf( "	-samp=		number of train input patterns in train file\n" );
	printf( "	-fwgt=		name of the file with weights (see documentation)\n" );
	printf( "	-use=		use the weight file or not (1, 0)\n" );
	printf( "	-fspec=		name of the file with initial parameters (see documentation)\n" );
	printf( "	-fout=		name of the file to write output to (see documentation)\n" );
	printf( "	-maxiter=	maximal number of iterations\n" );
	printf( "	-reps=		number of repetitions for each pattern before proceeding\n\n\n" );
	exit (1);
}

/* Parsing the input arguments gives flexibility: It is conveniant, 
   but also very usefull if the simulator has to be invoked many
   times. This function uses some global datastructure defined in 'cmd_pars.c'*/
void	parse( )
{
	int	cmd;
	char	rest[128];
	int	resti;
	long	restl;

	while ((cmd = cmdget( rest ))!= -1)
		{
		resti = atoi(rest);
		switch (cmd)
			{
			case CMD_INFILE:
				strcpy( tname, rest );
				break;
 			case CMD_SAMPLE:
				sample = resti;
				break;
 			case CMD_WGTS:
				strcpy( wgtname, rest );
				break;
 			case CMD_USEWGTS:
				use = resti;
				break;
 			case CMD_SPEC:
				strcpy( specname, rest );
				break;
 			case CMD_OUT:
				strcpy( specname, rest );
				break;
 			case CMD_ITER:
				maxiter = resti;
				break;
 			case CMD_REPS:
				rep = resti;
				break;
			case CMD_NULL:
				printf( "%s: unknown command [%s]\n", PROGNAME, rest );
				exit (2);
				break;
			}
		}
}

/* Reading the input vectors from a file into a structure */
void	get_pattern( tname, smpl)
	char	*tname;
	int	smpl;
{
	int	i, j, k, cnt, hlp;
	FILE	*ft;


	ft = fopen( tname, "r" );
	if (ft == NULL){
		error( NNFIERR );
	}

	if (sample <= 0){
		error( NN2FEWPATT );
	}

	/* 	Allocate input data structure in steps -
		I am of the opinion that to read the input data directly from a file
		into the U0 plane is not very elegant. I rather do this from another 
		specifically deicated input data structure.
		Desc:	The function returns a 3D array inputvect[sample][DIMINPUT][DIMINPUT] */
	if ( (inputvect = (double ***) malloc(sizeof(double **) * sample)) == NULL ){
		error( NNMALLOCINPERR );
	}
	
	for (i=0; i<sample; i++){
		if ( (inputvect[i] = (double **) malloc(sizeof(double *) * DIMINPUT) ) == NULL ){
			error( NNMALLOCINPERR );
		}
	}
	for (i=0; i<sample; i++){
		for (j=0; j<DIMINPUT; j++){
			if ( (inputvect[i][j] = (double *) malloc(sizeof(double) * DIMINPUT) ) == NULL ){
				error( NNMALLOCINPERR );
			}
		}
	}


	for (i=0; i<sample; i++){
		/*
		*	allocate input patterns
		*/
		hlp = 0;
		for (j=0; j<DIMINPUT; j++){
			for (k=0; k<DIMINPUT; k++){
				cnt = fscanf( ft, "%d", &hlp );
				inputvect[i][j][k] = (double) hlp;
				if (cnt < 1) {
					error( NNFIERR );
				}
			}
		}

	}

	fclose( ft );
}

/*	Reading weights for (the) excitatory connections between
	C_i-1 S_i+1 layers from a file directly into the network:
	The first number in the file is assumed to be the number
	of sets of weights starting from U0-U1, to read. The
	weights for each plane are assumed to be separated by a '.'
*/
void	get_weights( wgtname, net)
	char	*wgtname;
	NET	*net;
{
	char	separator=' ';
	int	i, j, k, l, m, hlp, cnt, dim;
	int	num_of_rcptflds=0, num_of_sets= -1;
	int	hlp1;
	int	dimLay, dimRfld; /* #planes, #wgts (have to be squared) */
	FILE	*ft;

	/*	complete the initialization:  This function basically runs over the network
		structures again and checks what has not been filled in and initializes
		accordingly. This concerns especially the fixed connections, the inhibitory
		connections and the variable connections not yet initialized.
		I want to go down to the plane level and check the init flag. */
	nn_init(net);

if (use == 1) {
	ft = fopen( wgtname, "r" );
	if (ft == NULL){
		error( NNFIERR );
	}

	/*cnt = fscanf( ft, "%d", &hlp );*/
	if ( fscanf( ft, "%d", &hlp ) < 1 ){
		error( NNFIERR );
	} else {
		if ( (hlp < 1) || (hlp > N_STAGES -1) ){
			error(NNWGTFIERR);
		} else {
			num_of_sets = hlp; /*	The count always starts at 0, no intermediate
						or interrrupted spec of wgts possible */
		}
	}


	for (i=1; i< num_of_sets + 1; i++){
		hlp = 0;
		num_of_rcptflds = NumRfld(Plane(LayerS(Stage(net,i)), 0)) - 1; /* max num of # rcptflds, which need to be initialized;
										Vs weights need not to be initialized */

		/* Read only the weights for the S planes not the V planes */
		for (j=0; j< DimLayer(LayerS(Stage(net, i+1))) - 1; j++){

			dim = DimRfld(Plane(LayerS(Stage(net, i+1)), 0));
			for (k=0; k< dim; k++){
				for (l=0; l< dim; l++){
					/* I assume the weights are integers-  they are going to be scaled to
						smaller doubles */
					if ( (cnt = fscanf( ft, "%d", &hlp1 )) < 1) error( NNFIERR );

			/*	read hlp1 into the (k, l)-wgt of the j-th plane of its m-th rcptfld-
				the same weight is put in the same position in all rcptflds of the plane*/
					for (m=0; m<num_of_rcptflds; m++){
						/* The patterns are given in integers. The 1s and 0s get scaled
						to smaller numbers */
					  Vi(vWeight(Plane(LayerS(Stage(net, i)), j), m), k, l) = 0.09 * (double) hlp1 + 0.01;
					}
				}
			}
			if ( fscanf( ft, "%s", &separator ) < 1 ) error( NNFIERR );
			if ( separator != '.' ) error(NNWGTFIERR);
			
			Plane(LayerS(Stage(net, i+1)), j)->init = 1; /* This plane has been initialized */
		}
	}

	/* Init was successful */
	fclose( ft );
}/*end 'use' */
}


/*	Reading parameters like the r_l's and q's from a file into global arrays:
	1. alpha, the parameter in the decaying exponential distribution (ref. McII, pg 205)
	2. the r_l's from stage 1 to stage (N_STAGES -1)
	3. q_l
	[3. beta (are not neccessary see McII, appendix, def in terms of phi() )]
	Notice, that there have to always be N_STAGES numbers in the file, even though
	the first or the last might make no sense. Set these numbers to zero.
*/
void	get_specs( specname )
	char	*specname;
{

	int	i, j, cnt;
	FILE	*ft;


	ft = fopen( specname, "r" );
	if (ft == NULL){
		error( NNFIERR );
	}

	/* scan alpha */
	if (fscanf( ft, "%lf", &alpha) < 1) error(NNPARERR);

	/* scan the r_l's */
	for (i=1; i<N_STAGES; i++){
 		if (fscanf( ft, "%lf", &r[i]) < 1) error(NNPARERR);
	/* scan the ql's */
	}
	for (i=1; i<N_STAGES; i++){
 		if (fscanf( ft, "%lf", &q[i]) < 1) error(NNPARERR);
	}

	/* Scan the coordinates of three planes which are going to be displayed */
	for (i=0; i<3; i++){
		
		fscanf( ft, "%lf", &disp[i].thresh);
		fscanf( ft, "%d", &disp[i].stage);
		fscanf( ft, "%d", &disp[i].layer);
		fscanf( ft, "%d", &disp[i].plane);
		fscanf( ft, "%d", &disp[i].flag);

	}

			
	fclose( ft );
}


/*	desc:	nn_init() initializes the weights of the network according to some specs-
		it update only those weights which have not already been initialized
		by a previous read-in from the wgtfile
		ref. KeI, pg460, 'a values' should depend on:
		fabs( (double)prev_plane/(double)dimS_prev -  (double)j/(double)dimS )
 		(double)k, (double)l, (double)j, (double)dimS 
		I will use random numbers for the time being, because the function
		above was not given */

int nn_init( net )
	NET	*net;
{
	int	num_of_stages, i, j, k, l, m, dim, dimS, dimC;
	int	num_of_Srcptflds, num_of_Crcptflds;
	double	hlp1=0.0;
	double	hlp2;
	FILE	*ft;


	srand(time(NULL));

	num_of_stages = DimNet(net);


	for (i=0; i< num_of_stages - 1; i++){ /* later we have to access the planes of the stage i+1 */

		/* max num of # rcptflds S layer*/
		num_of_Srcptflds = LayerS(Stage(net,i+1))->plane[0]->dimwgt;

		/* max num of # rcptflds C layer*/
		num_of_Crcptflds = LayerC(Stage(net,i+1))->plane[0]->dimwgt;

		dimS = DimLayer(LayerS(Stage(net, i+1)));
		dimC = DimLayer(LayerC(Stage(net, i+1)));

		/* first the S layer */
		for (j=0; j< dimS; j++){

			dim = DimRfld(Plane(LayerS(Stage(net, i+1)), j));
			for (k=0; k< dim; k++){
				for (l=0; l< dim; l++){

					/* SMALL NUMS ! & BiModal Weights to make 
					sure there will be an orientation */
					hlp1 = (double)(rand()%10001)/10001.0;
					if (hlp1 > 0.5){
						hlp1 = 0.1 + 0.01 * hlp1;
					} else {
						hlp1 = hlp1 * 0.001;
					}
					for (m=0; m<num_of_Srcptflds; m++){
						Vi(vWeight(Plane(LayerS(Stage(net, i+1)), j), m), k, l) = hlp1;
						hlp2 = Vi(vWeight(Plane(LayerS(Stage(net, i+1)), j), m), k, l);
					} /* m-loop */
				} /* l-loop */
			} /* k-loop */
		} /* j-loop */
	} /* i-loop */
return(0);
} /*End nn_init() */


/* Output function */

/* prototype if needed  */
/* void plane_plot( double, int, int, int, short ); */

void	plane_plot(net, which )
	/* if file_flag is TRUE output is to be written to a data file */

	NET	*net;
	int	which;

	{
	double	threshold;
	int	my_stage;
	int	my_sublayer;
	int	my_plane;
	int	file_flag;

	int x,y;
	int plane_size;
	FILE *p_file;
	p_file= fopen( outname, "a"); /* outname is global filename */

	/* make sure desired plane array is available globally !
	   this is because only one 2-D array size can be used by this function */
      
	threshold 	= disp[which].thresh;
	my_stage 	= disp[which].stage;
	my_sublayer 	= disp[which].layer;
	my_plane	= disp[which].plane;
	file_flag	= disp[which].flag;

	/* get size of desried plane */
        plane_size= DimPlane(net->stage[my_stage]->layerC->plane[my_plane]);
	
	printf("\nYou are in stage %d, layer %d, plane %d\n\n", my_stage, my_sublayer, my_plane);  

	if( file_flag )
	{
         fprintf(p_file, "\nYou are in stage %d, layer %d, plane %d\n\n", my_stage, my_sublayer, my_plane );
        } /* if loop */

	for( x=0; x<plane_size; x++ )
	{
  	  for( y=0; y<plane_size; y++ )
	  {
               /*
	      if( plane_array[x][y] >= threshold )  printf("*");
	      else			            printf(".");
               */

                if( !file_flag ) printf ("%10.8lf  ", net->stage[my_stage]->layerC->plane[my_plane]->unit[x][y].out );

		
		if( file_flag ) fprintf(p_file, "%10.8lf ", net->stage[my_stage]->layerC->plane[my_plane]->unit[x][y].out);

	      if( y==plane_size-1 && file_flag != TRUE )  printf("\n");
	      if( y==plane_size-1 && file_flag == TRUE ) fprintf(p_file, "\n" );    
	 
	  } /* y for loop */
	} /* x for loop */

} /* plane_plot() */
