/*
*----------------------------------------------------------------------------
*	file	: nnmath.c
*	desc	: Define all equations
*	by	: Frank Schnorrenberg
*	date	: 04-25-92
*
*	comment	:
*----------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "nntype.h"
#include "nnmath.h"

double	C_norm;

/*	desc:	The function s_cell_out() computes the out put of a s cell,
		given all necessary inputs */
double	s_cell_out(currlayer, prevlayer, st, k_l, x_orig, y_orig, x_rcptfld, y_rcptfld)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	int	st;		/* current stage index, I am not passing the pointer to the stage */
	int	k_l;		/* current plane index in current layer */
	/* Note: xorig and yorig are the coords in the CURRENT layer, they are NOT the coords
		of the center of the receptive field; for this an offset has to be computed,
		which is either 0 (same size planes) or 1 or 2 (this depends on the sizes of
		the previous and current plane; change of coordinate system)
	int	x_orig;
	int	y_orig;
	int	x_rcptfld;	/* x-pos of the current cell in the current layer */
	int	y_rcptfld;	/* y-pos of the current cell in the current layer */
{
	double	out, hlp1, hlp2, hlp3;


	/* Note: The x,y coordiantes are subject to a constraint: They have to correspond
		to some unit in the current plane, i.e. in a previous plane with bigger dim than
		the current plane x,y can only come from center positions (out of the
		coords corresponding to the image of the current plane on the previous
		plane */
	/*int	x_rcptfld;*/	/* x-pos of the receptive field center (a cells x index) */
	/*int	y_rcptfld;*/	/* y-pos of the receptive field center (a cells y index) */


	/* for the last stage the coords in previous and current layer are different */
	/* I do the job already in computeUstate()
	switch	(st){
		case 0:
			printf("Error: Stage index 0 should not occur here!");
			break;
		case 1:
			x_rcptfld = xorig;
			y_rcptfld = yorig;
			break;
		case 2:
			x_rcptfld = xorig + 1;
			y_rcptfld = yorig + 1;
			break;
		case 3:
			x_rcptfld = xorig + 1;
			y_rcptfld = yorig + 1;
			break;
		case 4:
			x_rcptfld = xorig + 2;
			y_rcptfld = yorig + 2;
			break;
		default:
			printf("Error: Invalid Stage index");
	}*/

	/* The position x_rcptfld, y_rcptfld is as explained above. For the s-cell it just means,
		it is the corresponding unit in some other layers plane e.g. the
		v layer, but in the other planes absolute coordiantes*/
	
	hlp1 = e(currlayer, prevlayer, k_l, x_rcptfld, y_rcptfld);
	hlp2 = h(currlayer, prevlayer, st, k_l, x_orig, y_orig, x_rcptfld, y_rcptfld);


	if (DEBUG == 1) {
		printf("unit(%d, %d), plane %d, stage %d has e()= %lf input\n", x_orig, y_orig, k_l, st, hlp1);
		printf("unit(%d, %d), plane %d, stage %d has h()= %lf input\n", x_orig, y_orig, k_l, st, hlp2);
	}


	out = 0.0;
	out = r[st] * phi( (1.0 + hlp1) / (1.0 + ( r[st]/(1.0 + r[st]) * hlp2)) - 1.0);
	/*out = r[st] * phi( (1.0 + hlp1) / (1.0 + (r[st] / (1.0 + r[st]) ) * hlp2 ) - 1.0);*/

	/*if (DEBUG == 1) {
		printf("out()= %lf\n", out);
	}*/


	return(out);

}


/*	desc: e() is used in s_cell_out(); e() gathers all excitatory connections */
double	e(currlayer, prevlayer, k_l, x, y)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	int	k_l;		/* the current plane index in the current layer */
	/* Note the change in coord system if the plane sizes of prev plane and current plane differ */
	int	x;		/* x is the first coord for the center position of the
				receptive field in the previous layer's planes */
	int	y;		/* y is the second coord */
{

	double	tmp;
	int	i, j;
	int	kappa, K_lprev;
	int	range;
	double	hlp1, hlp2;
	PLANE	*pl;
	
	/*	# of receptive fields, usually equal to # of planes in
		previous layer- this number does not change for the planes
		in one layer (in this implementation) */
	K_lprev	= NumRfld(Plane(currlayer, 0));

	/*	receptive field size of the units in current layer, which 
		look onto the previous layer planes; the size is equal for all
		planes in one layer */
	range	= DimRfld(Plane(currlayer, 0));
				/*	size of the receptive field:
					ATTENTION !!! The field is 2D, but the spec will
					be 1D. This means for the range in the formulas 
					we have to insert 2 loops over 'range' */

	/* for all planes in the previous layer */
	tmp = 0.0;
	for (kappa=0; kappa<K_lprev; kappa++){
		/*	receptive field is 2 dimensional & we do not need
			an absolute position ---> Sect2, 4, Appdx. in McIII;
			but we need an offset counter to access the output
			value of the correct C-cell Unit(Plane(prevlayer, kappa), x + i, y + j)
			where x, y are the absolute coords in the plane */

		/*	'range' has to be an odd number for this loop to work correctly:
			This is true for all cases in the net (receptive field size)*/
		for (i=((int) (-1 * (range/2))); i< ((int) (range/2) + 1); i++){
			for (j=((int) (-1 * (range/2))); j< ((int) (range/2) + 1); j++){			

				/* check for boundaries: we do not want to run over the array limits */
				if ( ((x + i) >= 0) && ((x + i) < DimPlane(Plane(prevlayer, kappa))) ) {
					if ( ((y + j) >= 0) && ((y + j) < DimPlane(Plane(prevlayer, kappa))) ) {

						/* 	'Go to the current layer plane (k_l), look up the
							correct receptive field (kappa), and access the right weight
							(i,j), Then pick the correct unit in the previous layers plane
							[(prevlayer, kappa). x+i, y+j]' */

		hlp1 = Vi(vWeight(Plane(currlayer, k_l), kappa), (i + (int)(range/2)), (j + (int)(range/2)));
		hlp2 = Unit(Plane(prevlayer, kappa), x + i, y + j).out;

		/*if ( (DEBUG == 1) && (K_lprev > 1) ){
			printf("Weight value: %lf\n", hlp1);
			printf("Out of Unit(Plane(prevlayer, %d), %d + %d, %d + %d).out = %lf\n", kappa, x, i, y, j, hlp2);
		}*/

		tmp += hlp1 * hlp2;
		/*if ( (DEBUG == 1) && (tmp != 0.0) ) printf("\007\007\n\nExcitation e() not equal zero: %lf\n\n", tmp);*/


					}
				}	

			}
		}
	}

	return(tmp);
}

/*	desc: h() is used in; h() gathers the inhibitory connections */

double	h(currlayer, prevlayer, stage, k_l, xorig, yorig, xrcptfld, yrcptfld)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	int	stage;	/* # of current stage */
	int	k_l;	/* current plane */
	int	xorig;	/* the coordinates in the current layer */
	int	yorig;
 	int	xrcptfld;	/* x is the first coord for the center position of the
				receptive field in the previous layer's planes */
	int	yrcptfld;	/* y is the second coordinate */
{
	/* ATTENTION:	The index K_lprev might be confusing. Technically we will go to the 
			k_lprev-layer, but we will only access inhibitoray cells, which have 
			their receptive field in k_lprev-layer, but they inhibit k_l-layer
			units according to their position in the plane only. So we are 
			accessing the last plane in the layer the current plane lies in, the V-cell plane.
 	*/
	double	tmp, b_l; /*	b_l is interpreted as the strength of the inhibitory
				connection; the values are saved in a datastructure 
				'inhib' in the plane (3D array) */
	double	u_vl; /* u_vl is the activation value of the V cell */
	int	K_l;

	K_l = DimLayer(currlayer);

	/* Want to access the V plane in the current layer; this is the plane with the highest index */
	b_l	= Plane(currlayer,  K_l - 1)->inhib[xorig][yorig][k_l];

	u_vl	= Unit(Plane(currlayer, K_l - 1), xorig, yorig).out;
	tmp =  (b_l * u_vl);

	return(tmp);
}


double	v_cell_out(currlayer, prevlayer, k_l, x, y)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	int	k_l;		/* the current plane index in the current layer */
	/* Note the change in coord system if the plane sizes of prev plane and current plane differ */
	int	x;		/* x is the first coord for the center position of the
				receptive field in the previous layer's planes */
	int	y;		/* y is the second coord */
{

	double	tmp, tmp1;
	int	i, j;
	int	kappa, K_lprev;
	int	range;
	
	/*	# of receptive fields, usually equal to # of planes in
		previous layer- this number does not change for the planes
		in one layer (in this implementation) */
 	K_lprev	= NumRfld(Plane(currlayer, 0));
	
	/*	receptive field size of the units in current layer, which 
		look onto the previous layer planes; the size is equal for all
		planes in one layer */
	range	= DimRfld(Plane(currlayer, 0));
				/*	size of the receptive field:
					ATTENTION !!! The field is 2D, but the spec will
					be 1D. This means for the range in the formulas 
					we have to insert 2 loops over 'range' */

 	/* for all planes in the previous layer */
	tmp = 0.0;
	tmp1 = 0.0;
	for (kappa=0; kappa<K_lprev; kappa++){
		/*	receptive field is 2 dimensional & we do not need
			an absolute position ---> Sect2, 4, Appdx. in McIII;
			but we need an offset counter to access the output
			value of the correct C-cell Unit(Plane(prevlayer, kappa), x + i, y + j)
			where x, y are the absolute coords in the plane */
		
		/*	'range' has to be an odd number for this loop to work correctly:
			This is true for all cases in the net (receptive field size)*/
		for (i=((int) (-1 * (range/2))); i< ((int) (range/2) + 1); i++){
			for (j=((int) (-1 * (range/2))); j< ((int) (range/2) + 1); j++){			

				/* check for boundaries: we do not want to run over the array limits */
				if ( ((x + i) >= 0) && ((x + i) < DimPlane(Plane(prevlayer, kappa))) ) {
					if ( ((y + j) >= 0) && ((y + j) < DimPlane(Plane(prevlayer, kappa))) ) {

				tmp1 += c(currlayer, prevlayer, i, j) * pow(Unit(Plane(prevlayer, kappa), x + i, y + j).out, 2.0);

					}
				}	

			}
		}

	tmp += sqrt(tmp1); /* This was the errro in the paper McII:
				First take the sqare root than add */
	
	}
	if (DEBUG_2 == 1){
		printf("Leaving v-cell_out()\n\n");
	}
	return(tmp);
}




/*	desc:	The function phi() computes a scalar value, based on
		a scalar input. phi(x , 0) = MAX(x, 0) */
double	phi(x)
	double	x;
{
	if (x >= 0.0) {
		return(x);
	} else {
		return((double)0.0);
	}
}


/*	desc:	The function c_cell_out() computes the out put of a c cell,
		given all necessary inputs */
/*	ATTENTION:	The inhibition for these units in this layer is not (yet) realized.
			I wonder if it is neccessary anyway. According to McIII, Apndx 
			eq. (A.3), the inhibitory connections do not exist. */
double	c_cell_out(currlayer, prevlayer, st, k_l, x_rcptfld, y_rcptfld)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	int	st;		/* current stage index, I am not passing the pointer to the stage */
	int	k_l;		/* current plane index in current layer */
	/* Note: xorig and yorig are the coords in the CURRENT layer, they are NOT the coords
		of the center of the receptive field; for this an offset has to be computed,
		which is either 0 (same size planes) or 2 (in the case U_c3 to U_s4, because
		the plane size is reduced from 7x7 to 3x3))*/
	int	x_rcptfld;	/* x-pos of the previous cell in the current layer */
	int	y_rcptfld;	/* y-pos of the previous cell in the current layer */
{
	double	out, tmp;
	int	i, j, range, kappa, num_rcptfld;
	/*int	x_rcptfld;*/	/* x-pos of the receptive field center (a cells x index) */
	/*int	y_rcptfld;*/	/* y-pos of the receptive field center (a cells y index) */

	range = DimRfld(Plane(currlayer, 0));

	/*The following job is done at the callers place !
	switch (st){
		case 1:
			x_rcptfld = xorig + 2;
			y_rcptfld = yorig + 2;
			break;
		case 2:
			x_rcptfld = xorig + 1;
			y_rcptfld = yorig + 1;
			break;
		case 3:
			x_rcptfld = xorig + 1;
			y_rcptfld = yorig + 1;
			break;
		case 4:
			x_rcptfld = xorig + 1;
			y_rcptfld = yorig + 1;
			break;
		default:
			printf("\n\nc_cell_out(): Wrong stage index %d\n\n", st);
	}*/

	tmp = 0.0;
	/* now, unlike for the S-cells, we do not do a loop over all previous planes
	but we just set kappa to number of receptive fields */
	
	if (st == 1){
		kappa = 0;
	} else {
		kappa = k_l;
	}

	/*	'range' has to be an odd number for this loop to work correctly:
		This is true for all cases in the net (receptive field size)*/
	for (i=((int) (-1 * (range/2))); i< ((int) (range/2) + 1); i++){
		for (j=((int) (-1 * (range/2))); j< ((int) (range/2) + 1); j++){		
			
			/* check for boundaries: we do not want to run over the array limits */
			if ( ((x_rcptfld + i) >= 0) && ((x_rcptfld + i) < DimPlane(Plane(prevlayer, kappa))) ) {
				if ( ((y_rcptfld + j) >= 0) && ((y_rcptfld + j) < DimPlane(Plane(prevlayer, kappa))) ) {

					/* 	'Go to the current layer plane (k_l), look up the
						correct receptive field (kappa), and access the right constant
						d(i, j).
						Then pick the correct unit in the previous layers plane
						[(prevlayer, kappa), x_rcptfld+i, y_rcptfld+j]' */

		/*if (DEBUG == 1) { 
			printf("Out of Unit(Plane(prevlayer, %d), %d + %d, %d + %d).out = %lf\n", 
			kappa, x_rcptfld, i, y_rcptfld, j, Unit(Plane(prevlayer, kappa), x_rcptfld + i, y_rcptfld + j).out);
		}*/

			tmp += d(i, j) * (Unit(Plane(prevlayer, kappa), x_rcptfld + i, y_rcptfld + j).out);
			/*if ( (DEBUG == 1) && (tmp != 0.0) ){
				printf("\007");
				printf("\n\nc_cell_out(): Stage %d, Layer %d, unit(%d, %d) = %lf\n\n", 
					st, kappa, x_rcptfld + i, y_rcptfld + j, tmp);
			}*/

				}
			}

		}/* END: receptive field */
	}/* END: receptive field */ 

	out = psi(tmp);

	if (DEBUG_2 == 1){
		printf("Leaving c-cell_out()\n\n");
	}
	return(out);

}

/* 	d() returns a factor reflecting strength of the fixed excitatory connections.
	Perhaps we want uniform strength to not artificially reduce the neighborhood!!! */
double	d(x, y)
	int	x, y;
{
	int	max;

	max = abs(x);
	if ( abs(y) > max ) max = abs(y);
	if (max == 0) return((double) 1.0);

	return(1.0/(double)(max * max));
}

/*double	d(x, y)
	int	x, y;
{
	int	max;

	max = abs(x);
	if ( abs(y) > max ) max = abs(y);
	if (max == 0) return((double) 1.0);

	return((double)1.0);*/ /* d is always 1.0*/
/*}*/



/*	desc:	The function psi() computes a scalar value depending on 
		a scalar x and an index l which stands for a stage;
		I use the def in McIII, appendix */
double	psi(x)
	double	x;
{

	if (x >= 0.0) {
		return( phi(x) / (1.0 + phi(x)) );
	} else {
		return((double)0.0);
	}
}



/*	desc:	The function c() computes a weight factor used e.g. in the inhibitory
		connections. The result also depends on a global parameter, called 
		alpha (ref: McII, pg. 205). The the function fills the array 'c_list'
		in the stage structure, which has exactly as many elements as the
		S layer has planes (one c value per plane).
		I implicitly assume that the size of the receptive field from one
		plane to the the previous planes does not change. Then I can use the c
		value just several times. */
double	c(currlayer, prevlayer, x, y)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	/* Note the change in coord system if the plane sizes of prev plane and current plane differ */
	int	x;		/* x is the x-offset from the center position*/
	int	y;		/* y is the y-offset from the center position*/
{
	double	tmp;
	double	norm_range;
	/*int	i, j;*/
	/*int	kappa;*/
	int	K_lprev;
	int	range;
	
	/*	# of receptive fields, usually equal to # of planes in
		previous layer- this number does not change for the planes
		in one layer (in this implementation) */
	K_lprev	= NumRfld(Plane(currlayer, 0));
	/*K_lprev	= prevlayer->plane[0]->dimwgt;*/

	/*	receptive field size of the units in current layer, which 
		look onto the previous layer planes; the size is equal for all
		planes in one layer */
	range	= DimRfld(Plane(currlayer, 0));
				/*	size of the receptive field:
					ATTENTION !!! The field is 2D, but the spec will
					be 1D. This means for the range in the formulas 
					we have to insert 2 loops over 'range' */

	norm_range = sqrt((double)(range * range));
	
	tmp = 0.0;
	/*	'range' has to be an odd number for this loop to work correctly:
		This is true for all cases in the net (receptive field size)*/
	/*for (i=((int) (-1 * (range/2))); i< ((int) (range/2) + 1); i++){
		for (j=((int) (-1 * (range/2))); j< ((int) (range/2) + 1); j++){*/

	/* tmp += pow(alpha, (sqrt((double)(i*i + j*j)) / norm_range));*/

	tmp = pow(alpha, (sqrt((double)(x*x + y*y)) / norm_range));

	/*	}
	}*/
	tmp = tmp/ C_norm;
	return(tmp);

}

double	cnorm(currlayer, prevlayer)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	/* Note the change in coord system if the plane sizes of prev plane and current plane differ */
{
	double	tmp, norm_dist;
	double	norm_range;
	int	i, j;
	int	kappa, K_lprev;
	int	range;
	
	/*	# of receptive fields, usually equal to # of planes in
		previous layer- this number does not change for the planes
		in one layer (in this implementation) */
	K_lprev	= NumRfld(Plane(currlayer, 0));
	/*K_lprev	= prevlayer->plane[0]->dimwgt;*/

	/*	receptive field size of the units in current layer, which 
		look onto the previous layer planes; the size is equal for all
		planes in one layer */
	range	= DimRfld(Plane(currlayer, 0));
				/*	size of the receptive field:
					ATTENTION !!! The field is 2D, but the spec will
					be 1D. This means for the range in the formulas 
					we have to insert 2 loops over 'range' */

	norm_range = sqrt((double)(range * range));
	
	/* for all planes in the previous layer */
	tmp = 0.0;
	C_norm = 0.0;
	for (kappa=0; kappa<K_lprev; kappa++){
		/*	'range' has to be an odd number for this loop to work correctly:
			This is true for all cases in the net (receptive field size)*/
		for (i=((int) (-1 * (range/2))); i< ((int) (range/2) + 1); i++){
			for (j=((int) (-1 * (range/2))); j< ((int) (range/2) + 1); j++){			

				/* Compute the normalizing constant */
				C_norm += pow(alpha, (sqrt((double)(i*i + j*j)) / norm_range));
			}
		}
	}
}


/* Update the variable weight values */
double	da(currlayer, prevlayer, st, k_l, xorig, yorig)
	LAYER	*currlayer;	/* pointer to current layer */
	LAYER	*prevlayer;	/* pointer to previous layer */
	/* Note the change in coord system if the plane sizes of prev plane and current plane differ */
	int	st;		/* current stage */
	int	k_l;		/* the current plane index in the current layer */
	int	xorig;
	int	yorig;
{
	double	tmp, tmp1, tmp2, tmp3;
	int	i, j;
	int	kappa, K_lprev;
	int	range;
	int	x_rcptfld, y_rcptfld;

	/*	# of receptive fields, usually equal to # of planes in
		previous layer- this number does not change for the planes
		in one layer (in this implementation) */
	K_lprev	= NumRfld(Plane(currlayer, 0));

	/*	receptive field size of the units in current layer, which 
		look onto the previous layer planes; the size is equal for all
		planes in one layer */
	range	= DimRfld(Plane(currlayer, 0));
				/*	size of the receptive field:
					ATTENTION !!! The field is 2D, but the spec will
					be 1D. This means for the range in the formulas 
					we have to insert 2 loops over 'range' */




	switch	(st){
		/* This is a coordiante system change to center the currplane within
		the prevplane */
		case 0:
			printf("Error: Stage index 0 should not occur here!");
			break;
		case 1:
			/*if ( (typ == 0) || (typ == 2) ){*/
				x_rcptfld = xorig;
				y_rcptfld = yorig;
			/*} else {
				x_rcptfld = xorig + 2;
				y_rcptfld = yorig + 2;
			}*/
			break;
		case 2:
			/*if ( (typ == 0) || (typ == 2) ){*/
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			/*} else {
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			}*/
			break;
		case 3:
			/*if ( (typ == 0) || (typ == 2) ){*/
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			/*} else {
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			}*/
			break;
		case 4:
			/*if ( (typ == 0) || (typ == 2) ){*/
				x_rcptfld = xorig + 2;
				y_rcptfld = yorig + 2;
			/*} else {
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			}*/
			break;
		default:
			printf("Error: Invalid Stage index");
	}



	/* for all planes in the previous layer */
	tmp = 0.0;
	/* cnorm(currlayer, prevlayer);*/
	for (kappa=0; kappa<K_lprev; kappa++){
		/*	receptive field is 2 dimensional & we do not need
			an absolute position ---> Sect2, 4, Appdx. in McIII;
			but we need an offset counter to access the output
			value of the correct C-cell Unit(Plane(prevlayer, kappa), x + i, y + j)
			where x, y are the absolute coords in the plane */

		/*	'range' has to be an odd number for this loop to work correctly:
			This is true for all cases in the net (receptive field size)*/
		for (i=((int) (-1 * (range/2))); i< ((int) (range/2) + 1); i++){
			for (j=((int) (-1 * (range/2))); j< ((int) (range/2) + 1); j++){			

				/* check for boundaries: we do not want to run over the array limits */
				if ( ((x_rcptfld + i) >= 0) && ((x_rcptfld + i) < DimPlane(Plane(prevlayer, kappa))) ) {
					if ( ((y_rcptfld + j) >= 0) && ((y_rcptfld + j) < DimPlane(Plane(prevlayer, kappa))) ) {

						/* 	'Go to the current layer plane (k_l), look up the
							correct receptive field (kappa), and access the right weight
							(i,j), Then pick the correct unit in the previous layers plane
							[(prevlayer, kappa). x+i, y+j]' */


		/*if (DEBUG == 1){
			tmp = Vi(vWeight(Plane(currlayer, k_l), kappa), (i + (int)(range/2)), (j + (int)(range/2)));
			printf("\nda(): Plane %d, RcptFld %d, Weight(%d, %d): %lf\n", k_l, kappa, (i + (int)(range/2)),
			(j + (int)(range/2)), tmp);
		}*/

		tmp1 = c(currlayer, prevlayer,i , j);
		tmp2 = c_cell_out(currlayer, prevlayer, st, k_l, x_rcptfld + i, y_rcptfld + j);

						tmp = q[st] * tmp1 * tmp2;
		Vi(vWeight(Plane(currlayer, k_l), kappa), (i + (int)(range/2)), (j + (int)(range/2))) += tmp;

		/*if ( (DEBUG == 1) && (tmp != 0.0) ){
			printf("\nWeight change: q[st] * c() * c_cell_out():\n	%lf =  %lf * %lf * %lf\n", tmp, q[st], tmp1, tmp2 );
		}*/



					}
				}	

			}
		}
	}

	return(tmp);
}

double	db(lyr, st, pl_idx, x, y)
	LAYER	*lyr;
	int	st;
	int	pl_idx;
	int	x;
	int	y;
{
	int	i, j;
	double	tmp;

	tmp = q[st] * Unit(Plane(lyr, (DimLayer(lyr) -1)), x, y).out;


	return(tmp);

}
 
