/*
*-----------------------------------------------------------------------------
*	file	: nntrain.c
*	desc	: Train the neocognitron in unsupervised mode
*	by	: Frank Schnorrenberg
*	date	: 04-15-92
*
*	comment	:
*
*-----------------------------------------------------------------------------
*/
#include <stdio.h>
#include <math.h>
#include <values.h>
#include <time.h>

#include "nntype.h"
#include "nnmath.h"
#include "nntrain.h"
#include "nnio.h"
#include "neov.h"
#include "nnerror.h"

extern	int	maxiter;
extern	int	rep;

struct{
	int	x;
	int	y;
	int	pl_idx;
	int	fill;
	} winners[DIM_MAXPLANE][DIM_MAXPLANE];
struct{
	int	x;
	int	y;
	int	pl_idx;
	int	fill;
	} winner_line[MAX_PLANES];


int	u_zero=0;/* This is a global variable (a flag)*/

static int	REPINTERVAL = 0;

/*
*=============================================================================
*	funct:	nn_train
*	dscpt:	train the neocognitron in unsupervised mode
*	given:	net = the neural net
*		inpvect = input vector array
*		np = number of patterns
*
*	retrn:	some value indicating success (?)
*=============================================================================
*/

double nn_train( net, inpvect)
NET	*net;
double ***inpvect;
{

	double	value=-1.0;
	/*int	cnt;*/

	/*cnt = 0;*/
	/*REPINTERVAL = report;
	*/

	value = nn_train1( net, inpvect);
 
	/* Here it would be easy to incorporate graphics commands to display
	sth. */

	/*if (report)
		nn_report(cnt, value);
 	}*/

	return(value);
}

/*
*=============================================================================
*	funct:	nn_report
*	dscpt:	print report lines to terminal
*	given:	cnt = number of train cycles
*
*	retrn:	nothing
*=============================================================================
*/
void nn_report( cnt, value )
int	cnt;
double	value;
{
	if (!(cnt%REPINTERVAL))
		{
		printf("\n%s: - trained %d cycles, value=%lf\n", PROGNAME, cnt, value);
		}
}

/*
*=============================================================================
*	funct:	nn_train1
*	dscpt:	Train net
*	given:	net 	= the neural net
*		inpvect = the input array
*		np 	= number of pattern
*		rep 	= number of repetions, before new pattern is presented
*	retrn:	value (Not really used as of now)
*=============================================================================
*/
double 
nn_train1(net, inpvect)
	NET            *net;
	double       ***inpvect;
{
	double          value=-1.0;
	int             i, j, samp, presentations;

	presentations = 0;
	while (maxiter > presentations){
		for (i = 0; i < sample; i++) {

			printf("\n\nCurrent Pattern Index: %d\n\n", i);

			for (j = 0; j < rep; j++) {

				printf("Pattern %d: currently repeating the %d time\n\n", i, j);

				/*
				 * Choose pattern: Can put in a function of the index
				 * i to choose e.g. pattern several times
				 */
				samp = i;

				/* Forward pass */
				nn_forward(net, inputvect, samp);

				/* 3 planes will be plotted accoding to the specs in 
				the spec file. We do it after the pass, because it 
				makes no difference- the network is fast enough */
				plane_plot(net, 0);
				plane_plot(net, 1);
				plane_plot(net, 2);

				/* Here we can insert the backward path */
				/* value = nn_backward( net, *(inpvect + i)); */

			}
		}
	presentations += sample * rep;
	}

	return (value);
}

/*
*=============================================================================
*	funct:	nn_forward
*	dscpt:	forward pass calculation
*	given:	net = the neural net
*		inpvect = the input pattern array
*	retrn:	nothing
*
*=============================================================================
*/
void nn_forward( net, inpvect, samp )
NET	*net;
double	***inpvect;
int	samp; /* what pattern to consider */
{
	STAGE	*I, *inputI;
	LAYER	**J, *inputJ, *prevLayer, *currLayer, *LastL;
	PLANE	*K, *inputK;
	UNIT	*L, *inputL;
	int	i, j, k, l1, l2, m, l, indx=-1, dimLay;
	double	sum, out, max=-1.0;
	int	type=-1;	/* 0 = 'S' layer, 1 = 'C' layer, 2 = 'SV' layer, 3 = 'SC' layer */

	/* Need to get an LAYER array with two elements */
	if (( J= (LAYER **)malloc(sizeof(LAYER *) * N_LAYERS_Rest) ) == NULL) {
		error(NNMALLOC_LAY);
		exit(2);
	};


/* Copy the input vector into the U0 stage (the C layer there) */
for (i=0; i<DIMINPUT; i++) {
	for (j=0; j<DIMINPUT; j++) {

	/* Out(Unit(Plane(LayerC(Stage(net, 0)), 0), i, j)) = inpvect[samp][i][j];*/
	Unit(Plane(LayerC(Stage(net, 0)), 0), i, j).out = inpvect[samp][i][j];

	}
}




/* 	For each stage i in the network, except stage 0, because this stage does not
	do anything except holding the input pattern*/
for (i=1; i<DimNet(net); i++) {
	/* Take a stage and make it the current one */
	I = Stage(net, i);
	
	if (DEBUG == 1){
		printf("nntrain.c: Number of stage is %d\n\n", i);
	}


	/* For each layer J in the network */
	for (j=0; j<DimStage(I); j++) {

	if (DEBUG == 1){
		printf("nntrain.c: Number of layer is %d\n\n", j);
	}


		/* Get the previous layer if there is one */
		if ( (DimStage(I) == 1) || (LayerC(Stage(net, i-1)) == NULL) || (LayerS(Stage(net, i-1)) == NULL) ){
			prevLayer =  NULL; /* No previous layer there, we are in stage 0 */
		} else {
			switch (j){
				case 0:
					prevLayer = LayerC(Stage(net, i-1));
					currLayer = LayerS(Stage(net, i));
					break;
				case 1:
					prevLayer = LayerS(Stage(net, i));
					currLayer = LayerC(Stage(net, i));
					break;
			}
		}
		
		/*	Take a layer and make it the current one.
			Need to account for 'binary' structure
			of the network: S layers, C layers and 
			for the U0 stage (C layer) 

		If we are in stage U0 the index=0 Layer is a C layer,
		otherwise index=0 implies S layer and index=1 implies
		C layer. To describe a layer without ambiguity the 
		stage index and the layer index are necessary */
		if ( (i != 0) && (j == 0) ) {
			/*J[j] = LayerS(I); *//* is eq. to '*(J + j) = ...' */
			type = 0;
 		} else {	
			/*J[j] = LayerC(I);*/
			type = 1;
		}

		/* Compute C_norm here, the normalization value for the c values */
		cnorm(currLayer, prevLayer);

		
	/*if (DEBUG == 1){
		printf("nntrain.c: Number of layer in numbers\n\n");
	}*/


		/* First compute the inhibition */
		/* For each unit in the INHIBITIONARY plane:
		This implementation uses only S plane cell inhibition! It
		is not hard to change. */
		k = DimLayer(currLayer) -1;
		K = currLayer->plane[k];
		if (type == 0){
			for (l1=0; l1<DimPlane(K); l1++) {
				for (l2=0; l2<DimPlane(K); l2++) {
			 		compute_Ustate(currLayer, prevLayer, 2, k, i, l1, l2);
				}
			}
		}



		/* For each plane K: Do not use the last plane in an S-layer, use it in an C-layer
			for the last plane in an S-Layer is the Vs-plane */
		if (type == 0) {
			dimLay = DimLayer(currLayer) - 1;
		} else {
			dimLay = DimLayer(currLayer);
		}
		for (k=0; k<dimLay; k++) {
		

			if (DEBUG == 1){
				printf("nntrain.c: Number of Plane is %d\n\n", k);
			}


			/* Take a plane and make it the current one */
			K = Plane(currLayer, i);
			/* I will use the 'long form to access everything under a plane, later I will use
				K for initialization of a pointer */


			/* For each unit in the plane */
			u_zero = 0;
			for (l1=0; l1<DimPlane(K); l1++) {
				for (l2=0; l2<DimPlane(K); l2++) {


					if (DEBUG_2 == 1){
						printf("nntrain.c: Coord of curr unit is  %d %d\n\n", l1, l2);
					}


					/* Take a unit and make it the current unit */
					/* L = Unit(K, l1, l2);  */
					/* L = &Unit(K, l1, l2); */

					/* 	Here we act upon the receptive fields:
 						Ultimately this has to be done for all units
						in all planes of the current layer. This is
						done by the enclosing loops. Then we have to
						find the winners of the competition and update
						only those winners. Then we proceed to the next
						layer. We take information on specific so to
						say 'effective' (used) dimensions of the structures
						from the 'dim' component in their structure.
						By this method it is easy to keep track of
						the number of receptive fields used */
					switch (type){
						case 0:
						  compute_Ustate(currLayer, prevLayer, type, k, i, l1, l2);
						  break;
						case 1:
						  compute_Ustate(currLayer, prevLayer, type, k, i, l1, l2);
						  break;
						default:
						  printf("\n\nError: Wrong layer type: %d\n\n", type);
					}

				}/*	END: For each unit, l2*/
			}/*	END: For each unit, l1*/


		/* Check the flag for all unit activation equal to 0 and beep */
		if (u_zero != 1) {
			printf("\007\007\n\nAll unit activation values are zero!\n\n");
		}

		}/*	END: For each plane */

		/*	Find the winning cells in all layers for all positions
			in the planes: Only one unit per position in all planes
			can win. We do NOT get neccessarily (# of units per plane)
			winning units, since first there is a competition between
			units in the same position but on different planes and then
			there is a competition between the cells on one plane.
			This competition determines ultimately which connections
			get updated. Nonetheless the other units stay active if they
			were active before. All cells receive of course inhibition 
			from the V cells, which co-determines the activation level. */

		/* Find the winners */

		/*	'winners' is a list of index pairs, identifying the winning
			units. The function find_winners consists of two loops:
			[1. Over all units; 2. Over all planes ]
			It's purpose is to find the unit with max. activation and append
			indices to the list. */
		/*if (DEBUG == 1){*/
		printf("Finding Winners in stage %d layer %d (0 = S, 1 = C)!\n\n", i, j);
		/*}*/
		find_winners(currLayer);
		/*	Is there more than one winning unit per plane? Choose the one
			with the highest activation and update the weights (if S layer) */
		trim_winners(currLayer);
		update_wgts(currLayer, prevLayer, i, type);

	}/*	END: For each layer */

} /*	END: For each stage */

/* Take a stage and make it the current one, the print some results */
I = Stage(net, 4);
 printf("The current pattern is pattern %d.\n\n", samp);
 max = -1.0;
 indx = -1;
 for(j=0; j< DimLayer(I->layerC); j++){
	for (k=0; k<1; k++){
		for (l=0; l<1; l++){
			if (Out(Unit(Plane(I->layerC, j), k, l)) > max){
				max = Out(Unit(Plane(I->layerC, j), k, l));
				indx = j;
			}
			printf("\nUnit %d, Last Layer has value %10.8lf\n", j, Out(Unit(Plane(I->layerC, j), k, l)) );
		}
	}
 }		
 printf("\nUnit %d has the maximal excitation (%10.8lf)\n\n", indx, max);

} /*	END: nn_forward */




int	compute_Ustate(currlayer, prevlayer, typ, k_l, st, xorig, yorig)
	LAYER	*currlayer;
	LAYER	*prevlayer;
	int	typ;
	/*PLANE	*pln;
	UNIT	*unt;*/
 	int	k_l;
	int	 st;
	int	xorig;
	int	yorig;
{
	int	x_rcptfld;
	int	y_rcptfld;
	double	ret_val;

	switch	(st){
		/* This is a coordiante system change to center the currplane within
		the prevplane */
		case 0:
			printf("Error: Stage index 0 should not occur here!");
			break;
		case 1:
			if ( (typ == 0) || (typ == 2) ){
				x_rcptfld = xorig;
				y_rcptfld = yorig;
			} else {
				x_rcptfld = xorig + 2;
				y_rcptfld = yorig + 2;
			}
			break;
		case 2:
			if ( (typ == 0) || (typ == 2) ){
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			} else {
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			}
			break;
		case 3:
			if ( (typ == 0) || (typ == 2) ){
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			} else {
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			}
			break;
		case 4:
			if ( (typ == 0) || (typ == 2) ){
				x_rcptfld = xorig + 2;
				y_rcptfld = yorig + 2;
			} else {
				x_rcptfld = xorig + 1;
				y_rcptfld = yorig + 1;
			}
			break;
		default:
			printf("Error: Invalid Stage index");
	}

	switch (typ){
		case 0:
			ret_val = s_cell_out(currlayer, prevlayer, st, k_l, xorig, yorig, x_rcptfld, y_rcptfld);
			currlayer->plane[k_l]->unit[xorig][yorig].out = ret_val;
			/*printf("Unit(%d, %d), inhib plane, stage %d in indicated has activation %lf now.\n\n",
				 x_rcptfld, y_rcptfld, st, ret_val);*/
			if (ret_val != 0.0) u_zero = 1;
			break;
		case 1:
			ret_val = c_cell_out(currlayer, prevlayer, st, k_l, x_rcptfld, y_rcptfld);
			currlayer->plane[k_l]->unit[xorig][yorig].out = ret_val;
			if (ret_val != 0.0) u_zero = 1;
			break;
		case 2:
			ret_val = v_cell_out(currlayer, prevlayer, (DimLayer(currlayer)-1), x_rcptfld, y_rcptfld);
			currlayer->plane[(DimLayer(currlayer)-1)]->unit[xorig][yorig].out = ret_val;
			if (ret_val != 0.0){
				u_zero = 1;
				/*printf("Unit(%d, %d), inhib plane, stage %d in indicated has activation %lf now.\n\n",
					x_rcptfld, y_rcptfld, st, ret_val);*/
			}
			break;
		default:
			printf("Error: Type was %d.\n\n", typ);
	}

return(0);
}


/*	desc: 	This fct finds the winners in the positional competition in
		a layer. Planes might occur several times */
		
int	find_winners(lyr)
	LAYER	*lyr;
{
	int	i, j, k;
	double	max=-100000.0;
	/* Note: The 'winners' array is global */

	/* init the winners array */
	for (j=0; j< DIM_MAXPLANE; j++){
		for (k=0; k< DIM_MAXPLANE; k++){
			winners[j][k].x 	= -1;
			winners[j][k].y 	= -1;
			winners[j][k].pl_idx 	= -1;
			winners[j][k].fill 	= -1;
		}
	}
	/* init the winner_line array */
	for (i=0; i< MAX_PLANES; i++){
		winner_line[i].x	= -1;
		winner_line[i].y	= -1;
		winner_line[i].pl_idx	= -1;
		winner_line[i].fill	= -1;
	}


	/* 	positional competition(NOT AMONG ALL, BUT ONLY THE C/S CELLS)
		The winners array will contain the winning unit, qualified by
		the x index and y index and the plane it came from. */
	/* for all planes in the layer */
	for (j=0; j<DimPlane(Plane(lyr, 0)); j++){
		for (k=0; k<DimPlane(Plane(lyr, 0)); k++){
			for (i=0; i<DimLayer(lyr) -1; i++){
			/* Note: The '-1' is to exclude the V plane */

				if (Unit(Plane(lyr, i), j, k).out > max){
					max = Unit(Plane(lyr, i), j, k).out;
					winners[j][k].x = j;
					winners[j][k].y = k;
					winners[j][k].pl_idx = i;
				}

			}
			winners[j][k].fill = DimPlane(Plane(lyr, 0));
			/* indicator for dim of plane used to fill the array */
		}
	}
return(0);
}


/*	desc:	This fct locates units in the winners array, with
		the same plane index and chooses the winners among
		units in one plane (intra-plane competition) */
int
trim_winners(lyr)
	LAYER          *lyr;
{
	int             i, j, k, l;
	int		pl_idx, pos_x, pos_y, curr_pl_idx;
	double		out_old, out_new;

	/* Note: The 'winners' array and the winner_line array are global */

	for (i = 0; i < MAX_PLANES; i++) {

		curr_pl_idx = i;

		for (j = 0; j < winners[0][0].fill; j++) {
			for (k = 0; k < winners[0][0].fill; k++) {

				if ( (winner_line[curr_pl_idx].pl_idx == -1) && (winners[j][k].pl_idx == curr_pl_idx) ) {
					winner_line[curr_pl_idx].pl_idx = curr_pl_idx;
					winner_line[curr_pl_idx].x = winners[j][k].x;
					winner_line[curr_pl_idx].y = winners[j][k].y;
					winner_line[curr_pl_idx].fill = 1;
				}

				if (winners[j][k].pl_idx == curr_pl_idx) {

				/* new candidate */
				 pl_idx	= curr_pl_idx;
				 pos_x	= winners[j][k].x;
				 pos_y	= winners[j][k].y;

				/* out_old = Unit(Plane(lyr, pl_idx), pos_x, pos_y).out; */
				 out_old = Unit(Plane(lyr, curr_pl_idx), j, k).out;
				 out_new = Unit(Plane(lyr, curr_pl_idx), pos_x, pos_y).out;

				 if ( out_new > out_old ) {

						winner_line[curr_pl_idx].pl_idx = curr_pl_idx;
						winner_line[curr_pl_idx].x 	= winners[j][k].x;
						winner_line[curr_pl_idx].y 	= winners[j][k].y;
						winner_line[curr_pl_idx].fill 	= 1;
					}
				}
			}
		}
	}
return(0);
}


int 
update_wgts(lyr, prelyr, st, type)
	LAYER		*lyr;
	LAYER		*prelyr;
	int		st;
	int		type;
{
	int	i, j, k;
	int	hlp1, hlp2, hlp3;

	/* Note: The 'winners' array and the 'winner_line' array are global */

	/* update the weights using the winning units from winner_line */
	for (i = 0; i < MAX_PLANES; i++) {
		if ( (winner_line[i].x >= 0) && (winner_line[i].y >= 0) ){
			hlp1 = winner_line[i].pl_idx;
			hlp2 = winner_line[i].x;
			hlp3 = winner_line[i].y;



			/* Only if S layer update inhibit and compute 'da' */
			if (type == 0) {
				/* Do the inhibitory weights all get updated always ??? if 
				yes this is an error here !!!!!*/
				lyr->plane[(DimLayer(lyr) - 1)]->inhib[hlp2][hlp3][hlp1] += db(lyr, st, hlp1, hlp2, hlp3);
				da(lyr, prelyr, st, hlp1, hlp2, hlp3);
				/* lyr->plane[hlp1]->unit[hlp2][hlp3].out = lyr->plane[hlp1]->unit[hlp2][hlp3].out; */

			} else {
				if (type == 1) {
					
				  lyr->plane[hlp1]->unit[hlp2][hlp3].out = c_cell_out(lyr, prelyr, st, hlp1, hlp2, hlp3);

				}
			}
		}
	}
	
return(0);
}


