/*
*----------------------------------------------------------------------------
*	file	: nntype.h
*	desc	: Define all types for the neural nets
*	by	: Frank Schnorrenberg
*	date	: 04-11-92
*----------------------------------------------------------------------------
*/

/*
*	trap multiple nntype include
*/
#ifndef	NNTYPE
#define	NNTYPE

#define	TRUE		1
#define	FALSE		0


#define DEBUG_1		0
#define	DEBUG_2		0
#define DEBUG		0

extern	int		u_zero;


#define FLAG_SIZE	10

/*	The construct 'DIMS' will be used to hold the original
	dimensions of the network */
typedef	int		DIMS;

/* 	The idea here is to provide one structure which is global
	and can be used to describe the 'state' of the network.
	What is basically provided is a global array for flaging
	exceptional conditions or the general system status */

typedef	struct	{
	int	*flags;
	DIMS	**dim_net;
	double	state;
	} STATE;


extern	STATE		state; /* not used */

typedef struct	{
	int	dim;		/* vector dimension */
	double	** vect;	/* 2D array, modelling the receptive field wgts */
	}	VECTOR;


typedef struct	{
	double	out;		/* output of unit */
	double	net;		/* net product */
	double	spec;		/* special for various usage */
	}	UNIT;

typedef struct	{
	int	init;		/*	indicator if plane (and its weights and units) have
					been initialized */

	int	dimunit;	/* 	number of units */
	UNIT	**unit;		/*	2D-array (!) of (pointers to) units(-structures) in a plane*/

	double	***inhib;	/*	This is a 3D array, indexed by the position
					of the unit to be inhibited (x, y) and the
					plane it resides in (pl_idx); only allocated for
					the last plane in a S-layer, which is considered to be
					the Vs plane(inhibiting the cells in the plane of its layer */

	int	dimwgt;		/* number of VECTOR structures attached */
	int	dim;		/* dimension of the receptive field allocated as a 2D array in
				VECTOR structures (this slot should be accessed in the VECTOR 
				structs but it is not)*/
	VECTOR	**wgtvect;	/* 1D array of VECTOR structures, as many as there are receptive
				fields for the plane (!). The weights are local to the plane
				and not to each unit. The special thing abt the units is that 
				they contribute their activation values, the weights all units
				have in common */
	}	PLANE;


typedef struct	{
	int	dim;		/* number of planes */
	PLANE	**plane;	/* array of planes in a layer */
	}	LAYER;



typedef struct	{
	int	dim;		/* number of layers */
	LAYER	*layerS;	/* S cell layer in a stage */
	LAYER	*layerC;	/* C cell layer in a stage */
	}	STAGE;



typedef	struct {
	int	dim;		/* number of stages */
	STAGE	**stage;	/* array of stages in the net */
	}	NET;



/* Building some terminology to access the datastructures */
#define	DimVect(vector)		((vector)->dim)
#define	Vect(vector)		((vector)->vect)
#define	Vi(vector,i, j)		((vector)->vect[i][j])

#define	Out(unit)		(unit.out)
#define	Net(unit)		(unit.net)	/* Not used */
#define	Spec(unit)		(unit.spec)	/* Not used */

#define	DimPlane(plane)		((plane)->dimunit)
#define	Unit(plane,i,j)		((plane)->unit[i][j])
#define	vWeight(plane, i)	((plane)->wgtvect[i])
#define	DimRfld(plane)		((plane)->dim)
#define	NumRfld(plane)		((plane)->dimwgt)

#define	DimLayer(layer)		((layer)->dim)
#define	Plane(layer,i)		((layer)->plane[i])

#define	DimStage(stage)		((stage)->dim)
#define	LayerS(stage)		((stage)->layerS)
#define	LayerC(stage)		((stage)->layerC)

#define	DimNet(nn)		((nn)->dim)
#define	Stage(nn,i)		((nn)->stage[i])



/* 	 
Example of how to use the macros:



 
*/



/*	The structure of the neocognitron is very much fixed, unlike
	in backpropagation, where it is usefull to be able to determine
	the network topology at the time when the program should be invoked.
	We hardwire the dimensionality of the vetors, planes,
	layers and stages. To change the structure it still should not be too bad
	since we use the above pointer structures. Be careful, not to expect too too
	much- especially when the plane sizes are changed in subsequent layers
	the programmer needs to adjust the mapping constants of coordinates from
	a plane in the one to a plane in the other (the earlier) plane, because
	I assume the planes always to be centered around each other:
	e.g. a 13x13 plane in the previous layer draws offset values of x=2 and
	y=2, if we assume the next planes are of size 9x9. The
	offset is necessary, since we have to use absolute coordinates of the previous
	layer's plane to refer to units in the previous layer's plane. The planes
	of subsequent layers should always be thought of as centered to each other.
	Then it should noit be difficult to determin the offsets:

			----------------------a plane of layer i
			|		      |
			|    -------------    |
			|   |		  |   |
			|   |		  |   | 
			|   | a plane of  |   | offset to adjust coords: x=2, y=2
			|   | layer i + 1 |   |	
			|   |		  |   |
			|    --------------   |
			|		      |
			-----------------------

*/


/* some parameters */
extern	double	r[];
extern	double	q[];
extern	double	alpha;

/* I/O stuff */
extern	int	sample;
extern	char	tname[128], wgtname[128], specname[128];
extern	double	***inputvect;
extern	int	maxiter;
extern	int	rep;
extern	int	use;

extern	double	C_norm;


/* 	Dimension of the input pattern is 19 by 19. The input pattern is read in
	row order into a 2D data structure. With this understanding we can
	specify: */

#define	DIMINPUT		19


/*	Number of stages (including the input stage, which only has one layer */

#define N_STAGES		5


/*	Dimension of Stage: How many layers? Except in the first stage
	there are always 2 layers */

#define	N_LAYERS_0		1
#define N_LAYERS_Rest		2


/*	Maximal Number of planes in the layers- this includes the VS and VC planes;
	Maximal Number of units in a plane (has to be squared to get true number) */

#define	MAX_PLANES		20
#define	DIM_MAXPLANE		19


/*	Number of planes in the Layer */
#define	N_PLANES_C0		1	/* Consider first layer a C layer */

#define	N_PLANES_S1		21	/* S-planes have one plane more: */
#define	N_PLANES_C1		20	/* The Vs plane */

#define	N_PLANES_S2		21
#define	N_PLANES_C2		20

#define	N_PLANES_S3		21
#define	N_PLANES_C3		20

#define	N_PLANES_S4		21
#define	N_PLANES_C4		20



/*	Size of planes/ Number of cells in each plane of each layer */

#define	S_PLANES_C0		19

#define	S_PLANES_S1		19
#define	S_PLANES_C1		15

#define	S_PLANES_S2		13
#define	S_PLANES_C2		11

#define	S_PLANES_S3		9
#define	S_PLANES_C3		7

#define	S_PLANES_S4		3
#define	S_PLANES_C4		1


/*	How many receptive fields per unit (i.e. plane):
	With # of receptive fields usually the number of previous layer cell planes
	is ment. In the case of C cells this might be different, since there is not
	this homogeneous one (C cell) to many (S cell receptive fields) connection
	scheme. In supervised learning, one decides e.g. about the connectivity and
	then trains accordingly.
	I like to make these differences explicit by introducing yet another set
	of constants. The constants are to be understood as maximum values, since
	in the C-layers the number varies (every unit is looking maximally at all
	previous planes). The variing number of receptive fields for C-layers is
	not implemented ('the j() fct for connectivity'), but could be done. In 
	the unsupervised case it makes no sense to me to have more than one 
	receptive field per C plane.*/
 

#define	Num_RcptFld_S0		0
#define	Num_RcptFld_C0		0

#define	Num_RcptFld_S1		1
#define	Num_RcptFld_C1		1

#define	Num_RcptFld_S2		20
#define	Num_RcptFld_C2		1

#define	Num_RcptFld_S3		20
#define	Num_RcptFld_C3		1

#define	Num_RcptFld_S4		20
#define	Num_RcptFld_C4		1





/*	Size of the receptive fields of the units in the planes
	in the S and C layers- the number has to be multiplied 
	by itself to get the number of weights, e.g. size 3 implies
	3 by 3 receptive field */

#define	DimWgtVec_S0		0	/* first S layer: non-existant */
#define	DimWgtVec_C0		0

#define	DimWgtVec_S1		5
#define	DimWgtVec_C1		5

#define	DimWgtVec_S2		5
#define	DimWgtVec_C2		5

#define	DimWgtVec_S3		5
#define	DimWgtVec_C3		5

#define	DimWgtVec_S4		5
#define	DimWgtVec_C4		3


#endif

